"use strict";

let cv, gl;

let program;

class Object {
	vertexBuffer;
	numVertices;
}

let o = new Object();

let lastFrame = Date.now();

let cameraPitch = 0.565;
let cameraYaw = 0.375;
let cameraDistance = 4;
let smoothCameraDistance = 4;

let mouseDragging = false;

let frameTimes = [];

function resizeCanvas() {
	const dpr = window.devicePixelRatio;
	cv.width = cv.parentElement.clientWidth * dpr;
	cv.height = cv.parentElement.clientHeight * dpr;
	gl.viewport(0, 0, cv.width, cv.height);
}

async function init() {
	cv = document.getElementById("cv");

	gl = cv.getContext("webgl");

	if (!gl) {
		window.alert("WebGL not supported");
	}

	const resizeObserver = new ResizeObserver(resizeCanvas);
	resizeObserver.observe(cv.parentElement);
	resizeCanvas();


	// input handling
	cv.addEventListener("mousedown", (e) => {
		if (e.button == 0) {
			mouseDragging = true;
		}
	});
	document.addEventListener("mouseup", (e) => {
		if (e.button == 0) {
			mouseDragging = false;
		}
	});
	document.addEventListener("mouseleave", (e) => {
		mouseDragging = false;
	});

	document.addEventListener("mousemove", (e) => {
		if (mouseDragging) {
			cameraYaw += e.movementX / 100;
			cameraPitch += e.movementY / 100;

			if (cameraPitch < -Math.PI / 2) cameraPitch = -Math.PI / 2;
			if (cameraPitch > Math.PI / 2) cameraPitch = Math.PI / 2;
		}
	});
	cv.addEventListener("wheel", (e) => {
		cameraDistance += e.deltaY / 100;
		if (cameraDistance < 1) cameraDistance = 1;
		if (cameraDistance > 15) cameraDistance = 15;
		e.preventDefault();
	});
	// end input handling

	console.log("compiling shaders...");
	program = await buildShaderProgram(gl, "shaders/vertex.glsl", "shaders/fragment.glsl");

	console.log("creating vertex buffer");


	let f = await readObjFile("obj/smooth_monkey.obj");
	o = new Object();

	o.vertexBuffer = gl.createBuffer();
	o.numVertices = f.length / 8;
	gl.bindBuffer(gl.ARRAY_BUFFER, o.vertexBuffer);
	gl.bufferData(gl.ARRAY_BUFFER, new Float32Array(f), gl.STATIC_DRAW);
	gl.bindBuffer(gl.ARRAY_BUFFER, null);


	// set clear colour
	gl.clearColor(...hexToRgb("181825"), 1.0);

	gl.enable(gl.DEPTH_TEST);
	gl.enable(gl.BLEND);
	gl.enable(gl.CULL_FACE);
	gl.frontFace(gl.CCW);
	gl.blendFunc(gl.SRC_ALPHA, gl.ONE_MINUS_SRC_ALPHA);

	// start drawing
	requestAnimationFrame(draw);
}

function setAttribPointers() {
	let positionAttribLocation = gl.getAttribLocation(program, "vertPosition");
	let normalAttribLocation = gl.getAttribLocation(program, "vertNormal");
	let textureAttribLocation = gl.getAttribLocation(program, "vertTexture");

	gl.vertexAttribPointer(positionAttribLocation, 3, gl.FLOAT, gl.FALSE, 8 * Float32Array.BYTES_PER_ELEMENT, 0);
	gl.vertexAttribPointer(textureAttribLocation, 2, gl.FLOAT, gl.FALSE, 8 * Float32Array.BYTES_PER_ELEMENT, 3 * Float32Array.BYTES_PER_ELEMENT);
	gl.vertexAttribPointer(normalAttribLocation, 3, gl.FLOAT, gl.FALSE, 8 * Float32Array.BYTES_PER_ELEMENT, 5 * Float32Array.BYTES_PER_ELEMENT);

	gl.enableVertexAttribArray(positionAttribLocation);
	gl.enableVertexAttribArray(textureAttribLocation);
	gl.enableVertexAttribArray(normalAttribLocation);
}

let frameCount = 0;
let lastStatUpdate = Date.now();

function updateStats(deltaTime) {
	frameCount++;

	let now = Date.now();
	let timeSinceLastUpdate = (now - lastStatUpdate) / 1000;

	if (timeSinceLastUpdate < 0.5) return;

	lastStatUpdate = now;

	let fpsDisplay = document.getElementById("fps");
	fpsDisplay.innerText = Math.round(frameCount / timeSinceLastUpdate)

	let msDisplay = document.getElementById("ms");
	if (frameTimes.length > 0) {
		msDisplay.innerText = (Math.round(frameTimes.reduce((prev, curr) => prev + curr) / frameTimes.length * 100) / 100) + "ms";
	}

	frameTimes = [];
	frameCount = 0;
}

let hue = 0;

async function draw() {
	let now = Date.now();
	let deltaTime = (now - lastFrame) / 1000;
	let frameStart = performance.now();

	hue += deltaTime / 5;
	if (hue > 1) hue = 0;

	smoothCameraDistance += (cameraDistance - smoothCameraDistance) / 8;

	updateStats(deltaTime);

	gl.clear(gl.COLOR_BUFFER_BIT | gl.DEPTH_BUFFER_BIT);

	let identity = [];
	mat4Identity(identity);

	// Real Matrices
	let projectionMatrix = [];
	mat4BuildPerspective(projectionMatrix, 90.0 / 180.0 * Math.PI, cv.width / cv.height, 0.1, 30);

	let viewMatrix = [];
	let cameraPosition = [Math.cos(cameraYaw) * smoothCameraDistance * Math.cos(cameraPitch), Math.sin(cameraPitch) * smoothCameraDistance, Math.sin(cameraYaw) * smoothCameraDistance * Math.cos(cameraPitch)];
	let cameraLookAt = [0, 0, 0];
	let cameraUp = [0, 1, 0];
	mat4BuildLookAt(viewMatrix, cameraPosition, cameraLookAt, cameraUp);

	let normalMatrix = [];
	mat3MakeNormal(normalMatrix, viewMatrix);

	gl.useProgram(program);
	gl.uniformMatrix4fv(gl.getUniformLocation(program, "projectionMatrix"), gl.FALSE, new Float32Array(projectionMatrix));
	gl.uniformMatrix4fv(gl.getUniformLocation(program, "modelViewMatrix"), gl.FALSE, new Float32Array(viewMatrix));
	gl.uniformMatrix3fv(gl.getUniformLocation(program, "normalMatrix"), gl.FALSE, new Float32Array(normalMatrix));

	let lightPosition = [Math.cos(hue * Math.PI * 2) * 3, 2.0, Math.sin(hue * Math.PI * 2) * 3, 1.0];
	mat4vec4Multiply(lightPosition, viewMatrix, lightPosition);

	gl.uniform3f(gl.getUniformLocation(program, "Light.position"), ...lightPosition.slice(0, 3));
	gl.uniform3f(gl.getUniformLocation(program, "Light.color"), 1, 1, 1);

	gl.uniform3f(gl.getUniformLocation(program, "Material.ambientColor"), 0.1, 0.1, 0.1);
	gl.uniform3f(gl.getUniformLocation(program, "Material.emitColor"), 0.05, 0.05, 0.0);
	gl.uniform3f(gl.getUniformLocation(program, "Material.diffColor"), 0.7, 0.8, 0.1);
	gl.uniform3f(gl.getUniformLocation(program, "Material.specColor"), 1, 1, 0.8);
	gl.uniform1f(gl.getUniformLocation(program, "Material.shininess"), 10);

	gl.bindBuffer(gl.ARRAY_BUFFER, o.vertexBuffer);
	setAttribPointers();
	gl.drawArrays(gl.TRIANGLES, 0, o.numVertices);

	frameTimes.push(performance.now() - frameStart);
	lastFrame = now;
	requestAnimationFrame(draw);
}

function hexToRgb(hex) {
	let r = parseInt(hex.substring(0, 2), 16) / 255;
	let g = parseInt(hex.substring(2, 4), 16) / 255;
	let b = parseInt(hex.substring(4, 6), 16) / 255;

	return [r, g, b];
}

function hslToRgb(h, s, l) {
	let r, g, b;

	if (s == 0) {
		r = g = b = l;
	} else {
		let hue2rgb = function hue2rgb(p, q, t) {
			if (t < 0) t += 1;
			if (t > 1) t -= 1;
			if (t < 1 / 6) return p + (q - p) * 6 * t;
			if (t < 1 / 2) return q;
			if (t < 2 / 3) return p + (q - p) * (2 / 3 - t) * 6;
			return p;
		}

		let q = l < 0.5 ? l * (1 + s) : l + s - l * s;
		let p = 2 * l - q;

		r = hue2rgb(p, q, h + 1 / 3);
		g = hue2rgb(p, q, h);
		b = hue2rgb(p, q, h - 1 / 3);
	}

	return [r, g, b];
}

window.addEventListener("DOMContentLoaded", init);